@@ -1,11 +1,9 @@
 # -*- test-case-name: twisted.web.test.test_xml -*-
-#
-# Copyright (c) 2001-2004 Twisted Matrix Laboratories.
+# Copyright (c) 2001-2008 Twisted Matrix Laboratories.
 # See LICENSE for details.
 
-#
-
-"""Micro Document Object Model: a partial DOM implementation with SUX.
+"""
+Micro Document Object Model: a partial DOM implementation with SUX.
 
 This is an implementation of what we consider to be the useful subset of the
 DOM.  The chief advantage of this library is that, not being burdened with
@@ -20,31 +18,38 @@ sample of XML.
 Microdom mainly focuses on working with HTML and XHTML.
 """
 
-from __future__ import nested_scopes
-
 # System Imports
-import copy, re
+import re
 from cStringIO import StringIO
 
-# Twisted Imports
-from twisted.web.sux import XMLParser, ParseError
-from twisted.python import reflect
-from twisted.python.util import InsensitiveDict
-
 # create NodeList class
 from types import ListType as NodeList
 from types import StringTypes, UnicodeType
-import sys
-dictsAreNotSequences=0
-try:
-    'x' in {}
-except TypeError:
-    dictsAreNotSequences = 1
+
+# Twisted Imports
+from twisted.web.sux import XMLParser, ParseError
+from twisted.python.util import InsensitiveDict
+
 
 def getElementsByTagName(iNode, name):
+    """
+    Return a list of all child elements of C{iNode} with a name matching
+    C{name}.
+
+    Note that this implementation does not conform to the DOM Level 1 Core
+    specification because it may return C{iNode}.
+
+    @param iNode: An element at which to begin searching.  If C{iNode} has a
+        name matching C{name}, it will be included in the result.
+
+    @param name: A C{str} giving the name of the elements to return.
+
+    @return: A C{list} of direct or indirect child elements of C{iNode} with
+        the name C{name}.  This may include C{iNode}.
+    """
     matches = []
     matches_append = matches.append # faster lookup. don't do this at home
-    slice=[iNode]
+    slice = [iNode]
     while len(slice)>0:
         c = slice.pop(0)
         if c.nodeName == name:
@@ -52,6 +57,8 @@ def getElementsByTagName(iNode, name):
         slice[:0] = c.childNodes
     return matches
 
+
+
 def getElementsByTagNameNoCase(iNode, name):
     name = name.lower()
     matches = []
@@ -96,7 +103,9 @@ class MismatchedTags(Exception):
         self.endCol) = filename, expect, got, begLine, begCol, endLine, endCol
 
     def __str__(self):
-        return "expected </%s>, got </%s> line: %s col: %s, began line: %s col: %s" % (self.expect, self.got, self.endLine, self.endCol, self.begLine, self.begCol)
+        return ("expected </%s>, got </%s> line: %s col: %s, began line: %s col: %s"
+                % (self.expect, self.got, self.endLine, self.endCol, self.begLine,
+                   self.begCol))
 
 
 class Node(object):
@@ -106,19 +115,28 @@ class Node(object):
         self.parentNode = parentNode
         self.childNodes = []
 
-    def __hash__(self):
-        return id(self)
+    def isEqualToNode(self, other):
+        """
+        Compare this node to C{other}.  If the nodes have the same number of
+        children and corresponding children are equal to each other, return
+        C{True}, otherwise return C{False}.
 
-    def isEqualToNode(self, n):
-        for a, b in zip(self.childNodes, n.childNodes):
+        @type other: L{Node}
+        @rtype: C{bool}
+        """
+        if len(self.childNodes) != len(other.childNodes):
+            return False
+        for a, b in zip(self.childNodes, other.childNodes):
             if not a.isEqualToNode(b):
-                return 0
-        return 1
+                return False
+        return True
 
-    def writexml(self, stream, indent='', addindent='', newl='', strip=0, nsprefixes={}, namespace=''):
+    def writexml(self, stream, indent='', addindent='', newl='', strip=0,
+                 nsprefixes={}, namespace=''):
         raise NotImplementedError()
 
-    def toxml(self, indent='', addindent='', newl='', strip=0, nsprefixes={}, namespace=''):
+    def toxml(self, indent='', addindent='', newl='', strip=0, nsprefixes={},
+              namespace=''):
         s = StringIO()
         self.writexml(s, indent, addindent, newl, strip, nsprefixes, namespace)
         rv = s.getvalue()
@@ -160,7 +178,9 @@ class Node(object):
         assert isinstance(newChild, Node)
         #if newChild.parentNode:
         #    newChild.parentNode.removeChild(newChild)
-        assert oldChild.parentNode is self, 'oldChild (%s): oldChild.parentNode (%s) != self (%s)' % (oldChild, oldChild.parentNode, self)
+        assert (oldChild.parentNode is self,
+                ('oldChild (%s): oldChild.parentNode (%s) != self (%s)'
+                 % (oldChild, oldChild.parentNode, self)))
         self.childNodes[self.childNodes.index(oldChild)] = newChild
         oldChild.parentNode = None
         newChild.parentNode = self
@@ -205,7 +225,8 @@ class Document(Node):
     doctype = None
 
     def isEqualToDocument(self, n):
-        return (self.doctype == n.doctype) and self.isEqualToNode(n)
+        return (self.doctype == n.doctype) and Node.isEqualToNode(self, n)
+    isEqualToNode = isEqualToDocument
 
     def get_documentElement(self):
         return self.childNodes[0]
@@ -215,11 +236,13 @@ class Document(Node):
         assert not self.childNodes, "Only one element per document."
         Node.appendChild(self, c)
 
-    def writexml(self, stream, indent='', addindent='', newl='', strip=0, nsprefixes={}, namespace=''):
+    def writexml(self, stream, indent='', addindent='', newl='', strip=0,
+                 nsprefixes={}, namespace=''):
         stream.write('<?xml version="1.0"?>' + newl)
         if self.doctype:
             stream.write("<!DOCTYPE "+self.doctype+">" + newl)
-        self.documentElement.writexml(stream, indent, addindent, newl, strip, nsprefixes, namespace)
+        self.documentElement.writexml(stream, indent, addindent, newl, strip,
+                                      nsprefixes, namespace)
 
     # of dubious utility (?)
     def createElement(self, name, **kw):
@@ -257,8 +280,10 @@ class EntityReference(Node):
         if not isinstance(n, EntityReference):
             return 0
         return (self.eref == n.eref) and (self.nodeValue == n.nodeValue)
+    isEqualToNode = isEqualToEntityReference
 
-    def writexml(self, stream, indent='', addindent='', newl='', strip=0, nsprefixes={}, namespace=''):
+    def writexml(self, stream, indent='', addindent='', newl='', strip=0,
+                 nsprefixes={}, namespace=''):
         stream.write(self.nodeValue)
 
     def cloneNode(self, deep=0, parent=None):
@@ -273,12 +298,14 @@ class CharacterData(Node):
 
     def isEqualToCharacterData(self, n):
         return self.value == n.value
+    isEqualToNode = isEqualToCharacterData
 
 
 class Comment(CharacterData):
     """A comment node."""
 
-    def writexml(self, stream, indent='', addindent='', newl='', strip=0, nsprefixes={}, namespace=''):
+    def writexml(self, stream, indent='', addindent='', newl='', strip=0,
+                 nsprefixes={}, namespace=''):
         val=self.data
         if isinstance(val, UnicodeType):
             val=val.encode('utf8')
@@ -294,10 +321,22 @@ class Text(CharacterData):
         CharacterData.__init__(self, data, parentNode)
         self.raw = raw
 
+
+    def isEqualToNode(self, other):
+        """
+        Compare this text to C{text}.  If the underlying values and the C{raw}
+        flag are the same, return C{True}, otherwise return C{False}.
+        """
+        return (
+            CharacterData.isEqualToNode(self, other) and
+            self.raw == other.raw)
+
+
     def cloneNode(self, deep=0, parent=None):
         return Text(self.nodeValue, parent, self.raw)
 
-    def writexml(self, stream, indent='', addindent='', newl='', strip=0, nsprefixes={}, namespace=''):
+    def writexml(self, stream, indent='', addindent='', newl='', strip=0,
+                 nsprefixes={}, namespace=''):
         if self.raw:
             val = self.nodeValue
             if not isinstance(val, StringTypes):
@@ -321,7 +360,8 @@ class CDATASection(CharacterData):
     def cloneNode(self, deep=0, parent=None):
         return CDATASection(self.nodeValue, parent)
 
-    def writexml(self, stream, indent='', addindent='', newl='', strip=0, nsprefixes={}, namespace=''):
+    def writexml(self, stream, indent='', addindent='', newl='', strip=0,
+                 nsprefixes={}, namespace=''):
         stream.write("<![CDATA[")
         stream.write(self.nodeValue)
         stream.write("]]>")
@@ -341,12 +381,11 @@ class Element(Node):
     preserveCase = 0
     caseInsensitive = 1
     nsprefixes = None
-    namespace = ''
 
     def __init__(self, tagName, attributes=None, parentNode=None,
-                        filename=None, markpos=None,
-                        caseInsensitive=1, preserveCase=0,
-                 namespace=''):
+                 filename=None, markpos=None,
+                 caseInsensitive=1, preserveCase=0,
+                 namespace=None):
         Node.__init__(self, parentNode)
         self.preserveCase = preserveCase or not caseInsensitive
         self.caseInsensitive = caseInsensitive
@@ -360,7 +399,7 @@ class Element(Node):
                 self.attributes[k] = unescape(v)
 
         if caseInsensitive:
-            self.attributes = InsensitiveDict(self.attributes, 
+            self.attributes = InsensitiveDict(self.attributes,
                                               preserve=preserveCase)
 
         self.endTagName = self.nodeName = self.tagName = tagName
@@ -381,9 +420,24 @@ class Element(Node):
 
     def isEqualToElement(self, n):
         if self.caseInsensitive:
-            return (self.attributes == n.attributes) and (self.nodeName.lower() == n.nodeName.lower())
+            return ((self.attributes == n.attributes)
+                    and (self.nodeName.lower() == n.nodeName.lower()))
         return (self.attributes == n.attributes) and (self.nodeName == n.nodeName)
 
+
+    def isEqualToNode(self, other):
+        """
+        Compare this element to C{other}.  If the C{nodeName}, C{namespace},
+        C{attributes}, and C{childNodes} are all the same, return C{True},
+        otherwise return C{False}.
+        """
+        return (
+            self.nodeName.lower() == other.nodeName.lower() and
+            self.namespace == other.namespace and
+            self.attributes == other.attributes and
+            Node.isEqualToNode(self, other))
+
+
     def cloneNode(self, deep=0, parent=None):
         clone = Element(
             self.tagName, parentNode=parent, namespace=self.namespace,
@@ -424,21 +478,11 @@ class Element(Node):
         if name in self.attributes:
             del self.attributes[name]
 
-    def removeAttribute_has_key(self, name):
-        if self.attributes.has_key(name):
-            del self.attributes[name]
-
     def hasAttribute(self, name):
         return name in self.attributes
 
-    def hasAttribute_has_key(self, name):
-        return self.attributes.has_key(name)
-
-    if dictsAreNotSequences:
-        hasAttribute = hasAttribute_has_key
-        removeAttribute = removeAttribute_has_key
-
-    def writexml(self, stream, indent='', addindent='', newl='', strip=0, nsprefixes={}, namespace=''):
+    def writexml(self, stream, indent='', addindent='', newl='', strip=0,
+                 nsprefixes={}, namespace=''):
         # write beginning
         ALLOWSINGLETON = ('img', 'br', 'hr', 'base', 'meta', 'link', 'param',
                           'area', 'input', 'col', 'basefont', 'isindex',
@@ -458,7 +502,8 @@ class Element(Node):
         if self.nsprefixes:
             newprefixes = self.nsprefixes.copy()
             for ns in nsprefixes.keys():
-                del newprefixes[ns]
+                if ns in newprefixes:
+                    del newprefixes[ns]
         else:
              newprefixes = {}
 
@@ -467,7 +512,7 @@ class Element(Node):
             begin = [newl, indent] + begin
         bext = begin.extend
         writeattr = lambda _atr, _val: bext((' ', _atr, '="', escape(_val), '"'))
-        if namespace != self.namespace and self.namespace:
+        if namespace != self.namespace and self.namespace is not None:
             if nsprefixes.has_key(self.namespace):
                 prefix = nsprefixes[self.namespace]
                 bext(prefix+':'+self.tagName)
@@ -506,11 +551,12 @@ class Element(Node):
                 if self.tagName in BLOCKELEMENTS and \
                    self.tagName in FORMATNICELY:
                     w(j((newl, newindent)))
-                child.writexml(stream, newindent, addindent, newl, strip, downprefixes, self.namespace)
+                child.writexml(stream, newindent, addindent, newl, strip,
+                               downprefixes, self.namespace)
             if self.tagName in BLOCKELEMENTS:
                 w(j((newl, indent)))
             w(j(("</", self.endTagName, '>')))
-            
+
         elif self.tagName.lower() not in ALLOWSINGLETON:
             w(j(('></', self.endTagName, '>')))
         else:
@@ -581,9 +627,10 @@ class MicroDOMParser(XMLParser):
                     }
 
 
-    def __init__(self, beExtremelyLenient=0, caseInsensitive=1, preserveCase=0, soonClosers=soonClosers, laterClosers=laterClosers):
+    def __init__(self, beExtremelyLenient=0, caseInsensitive=1, preserveCase=0,
+                 soonClosers=soonClosers, laterClosers=laterClosers):
         self.elementstack = []
-        d = {'xmlns': 'xmlns', '':''}
+        d = {'xmlns': 'xmlns', '': None}
         dr = _reverseDict(d)
         self.nsstack = [(d,None,dr)]
         self.documents = []
@@ -609,7 +656,7 @@ class MicroDOMParser(XMLParser):
             return None
 
     COMMENT = re.compile(r"\s*/[/*]\s*")
-    
+
     def _fixScriptElement(self, el):
         # this deals with case where there is comment or CDATA inside
         # <script> tag and we want to do the right thing with it
@@ -628,7 +675,7 @@ class MicroDOMParser(XMLParser):
             if match:
                 prefix = match.group()
                 oldvalue = oldvalue[len(prefix):]
-            
+
             # now see if contents are actual node and comment or CDATA
             try:
                 e = parseString("<a>%s</a>" % oldvalue).childNodes[0]
@@ -642,7 +689,7 @@ class MicroDOMParser(XMLParser):
                 if prefix:
                     el.childNodes.append(Text(prefix))
                 el.childNodes.append(e)
-    
+
     def gotDoctype(self, doctype):
         self._mddoctype = doctype
 
@@ -687,7 +734,7 @@ class MicroDOMParser(XMLParser):
                      namespace=namespaces.get(''))
         revspaces = _reverseDict(newspaces)
         el.addPrefixes(revspaces)
-        
+
         if newspaces:
             rscopy = self.nsstack[-1][2].copy()
             rscopy.update(revspaces)
@@ -775,8 +822,9 @@ class MicroDOMParser(XMLParser):
             self.documents.append(el)
         if self.beExtremelyLenient and el.tagName == "script":
             self._fixScriptElement(el)
-    
+
     def connectionLost(self, reason):
+        XMLParser.connectionLost(self, reason) # This can cause more events!
         if self.elementstack:
             if self.beExtremelyLenient:
                 self.documents.append(self.elementstack[0])
@@ -809,17 +857,14 @@ def parse(readable, *args, **kwargs):
     if mdp.beExtremelyLenient:
         if len(mdp.documents) == 1:
             d = mdp.documents[0]
+            if not isinstance(d, Element):
+                el = Element("html")
+                el.appendChild(d)
+                d = el
         else:
-            d = None
-            for el in mdp.documents:
-                if isinstance(el, Element):
-                    if d is not None:
-                        d = None
-                        break
-                    d = el
-            if d is None:
-                d = Element("html")
-                d.childNodes[:] = mdp.documents
+            d = Element("html")
+            for child in mdp.documents:
+                d.appendChild(child)
     else:
         d = mdp.documents[0]
     doc = Document(d)
