@@ -1,31 +1,31 @@
-# Copyright (c) 2001-2004 Twisted Matrix Laboratories.
+# Copyright (c) 2001-2008 Twisted Matrix Laboratories.
 # See LICENSE for details.
 
+"""
+Tests for L{twisted.web.client}.
+"""
+
+import os
+
+from urlparse import urlparse
 
 from twisted.trial import unittest
 from twisted.web import server, static, client, error, util, resource
 from twisted.internet import reactor, defer, interfaces
-from twisted.python.util import sibpath
-from twisted.python import components
+from twisted.python.filepath import FilePath
+from twisted.protocols.policies import WrappingFactory
+from twisted.test.proto_helpers import StringTransport
 
 try:
     from twisted.internet import ssl
 except:
     ssl = None
 
-import os
 
-serverCallID = None
-
-class LongTimeTakingResource(resource.Resource):
+class ForeverTakingResource(resource.Resource):
     def render(self, request):
-        global serverCallID
-        serverCallID =  reactor.callLater(1, self.writeIt, request)
         return server.NOT_DONE_YET
 
-    def writeIt(self, request):
-        request.write("hello!!!")
-        request.finish()
 
 class CookieMirrorResource(resource.Resource):
     def render(self, request):
@@ -51,20 +51,34 @@ class NoLengthResource(resource.Resource):
 
     def render(self, request):
         return "nolength"
-    
-class HostHeaderResource(resource.Resource):
 
+
+
+class HostHeaderResource(resource.Resource):
+    """
+    A testing resource which renders itself as the value of the host header
+    from the request.
+    """
     def render(self, request):
-        return request.received_headers["host"]
+        return request.received_headers['host']
+
+
 
 class PayloadResource(resource.Resource):
-
+    """
+    A testing resource which renders itself as the contents of the request body
+    as long as the request body is 100 bytes long, otherwise which renders
+    itself as C{"ERROR"}.
+    """
     def render(self, request):
         data = request.content.read()
-        if len(data) != 100 or int(request.received_headers["content-length"]) != 100:
+        contentLength = request.received_headers['content-length']
+        if len(data) != 100 or int(contentLength) != 100:
             return "ERROR"
         return data
 
+
+
 class BrokenDownloadResource(resource.Resource):
 
     def render(self, request):
@@ -73,32 +87,129 @@ class BrokenDownloadResource(resource.Resource):
         request.write('abc')
         return ''
 
+class CountingRedirect(util.Redirect):
+    """
+    A L{util.Redirect} resource that keeps track of the number of times the
+    resource has been accessed.
+    """
+    def __init__(self, *a, **kw):
+        util.Redirect.__init__(self, *a, **kw)
+        self.count = 0
+
+    def render(self, request):
+        self.count += 1
+        return util.Redirect.render(self, request)
+
+
+
+class ParseUrlTestCase(unittest.TestCase):
+    """
+    Test URL parsing facility and defaults values.
+    """
+
+    def testParse(self):
+        scheme, host, port, path = client._parse("http://127.0.0.1/")
+        self.assertEquals(path, "/")
+        self.assertEquals(port, 80)
+        scheme, host, port, path = client._parse("https://127.0.0.1/")
+        self.assertEquals(path, "/")
+        self.assertEquals(port, 443)
+        scheme, host, port, path = client._parse("http://spam:12345/")
+        self.assertEquals(port, 12345)
+        scheme, host, port, path = client._parse("http://foo ")
+        self.assertEquals(host, "foo")
+        self.assertEquals(path, "/")
+        scheme, host, port, path = client._parse("http://egg:7890")
+        self.assertEquals(port, 7890)
+        self.assertEquals(host, "egg")
+        self.assertEquals(path, "/")
+
+
+    def test_externalUnicodeInterference(self):
+        """
+        L{client._parse} should return C{str} for the scheme, host, and path
+        elements of its return tuple, even when passed an URL which has
+        previously been passed to L{urlparse} as a C{unicode} string.
+        """
+        badInput = u'http://example.com/path'
+        goodInput = badInput.encode('ascii')
+        urlparse(badInput)
+        scheme, host, port, path = client._parse(goodInput)
+        self.assertTrue(isinstance(scheme, str))
+        self.assertTrue(isinstance(host, str))
+        self.assertTrue(isinstance(path, str))
+
+
+
+class HTTPPageGetterTests(unittest.TestCase):
+    """
+    Tests for L{HTTPPagerGetter}, the HTTP client protocol implementation
+    used to implement L{getPage}.
+    """
+    def test_earlyHeaders(self):
+        """
+        When a connection is made, L{HTTPPagerGetter} sends the headers from
+        its factory's C{headers} dict.  If I{Host} or I{Content-Length} is
+        present in this dict, the values are not sent, since they are sent with
+        special values before the C{headers} dict is processed.  If
+        I{User-Agent} is present in the dict, it overrides the value of the
+        C{agent} attribute of the factory.  If I{Cookie} is present in the
+        dict, its value is added to the values from the factory's C{cookies}
+        attribute.
+        """
+        factory = client.HTTPClientFactory(
+            'http://foo/bar',
+            agent="foobar",
+            cookies={'baz': 'quux'},
+            postdata="some data",
+            headers={
+                'Host': 'example.net',
+                'User-Agent': 'fooble',
+                'Cookie': 'blah blah',
+                'Content-Length': '12981',
+                'Useful': 'value'})
+        transport = StringTransport()
+        protocol = client.HTTPPageGetter()
+        protocol.factory = factory
+        protocol.makeConnection(transport)
+        self.assertEqual(
+            transport.value(),
+            "GET /bar HTTP/1.0\r\n"
+            "Host: example.net\r\n"
+            "User-Agent: foobar\r\n"
+            "Content-Length: 9\r\n"
+            "Useful: value\r\n"
+            "connection: close\r\n"
+            "Cookie: blah blah; baz=quux\r\n"
+            "\r\n"
+            "some data")
+
+
+
 class WebClientTestCase(unittest.TestCase):
     def _listen(self, site):
         return reactor.listenTCP(0, site, interface="127.0.0.1")
 
     def setUp(self):
-        name = str(id(self)) + "_webclient"
-        if not os.path.exists(name):
-            os.mkdir(name)
-            f = open(os.path.join(name, "file"), "wb")
-            f.write("0123456789")
-            f.close()
+        name = self.mktemp()
+        os.mkdir(name)
+        FilePath(name).child("file").setContent("0123456789")
         r = static.File(name)
         r.putChild("redirect", util.Redirect("/file"))
-        r.putChild("wait", LongTimeTakingResource())
+        self.infiniteRedirectResource = CountingRedirect("/infiniteRedirect")
+        r.putChild("infiniteRedirect", self.infiniteRedirectResource)
+        r.putChild("wait", ForeverTakingResource())
         r.putChild("error", ErrorResource())
         r.putChild("nolength", NoLengthResource())
         r.putChild("host", HostHeaderResource())
         r.putChild("payload", PayloadResource())
         r.putChild("broken", BrokenDownloadResource())
-        site = server.Site(r, timeout=None)
-        self.port = self._listen(site)
+        self.site = server.Site(r, timeout=None)
+        self.wrapper = WrappingFactory(self.site)
+        self.port = self._listen(self.wrapper)
         self.portno = self.port.getHost().port
 
     def tearDown(self):
-        if serverCallID and serverCallID.active():
-            serverCallID.cancel()
         return self.port.stopListening()
 
     def getURL(self, path):
@@ -116,28 +227,70 @@ class WebClientTestCase(unittest.TestCase):
         d = self.assertFailure(d, client.PartialDownloadError)
         d.addCallback(lambda exc: self.assertEquals(exc.response, "abc"))
         return d
-    
+
     def testHostHeader(self):
         # if we pass Host header explicitly, it should be used, otherwise
         # it should extract from url
         return defer.gatherResults([
             client.getPage(self.getURL("host")).addCallback(self.assertEquals, "127.0.0.1"),
             client.getPage(self.getURL("host"), headers={"Host": "www.example.com"}).addCallback(self.assertEquals, "www.example.com")])
-    
-    def testGetPage(self):
-        return client.getPage(self.getURL("file")).addCallback(self.assertEquals, "0123456789")
 
-    def testTimeoutNotTriggering(self):
-        # Test that when the timeout doesn't trigger, things work as expected.
-        d = client.getPage(self.getURL("wait"), timeout=100)
-        d.addCallback(self.assertEquals, "hello!!!")
+
+    def test_getPage(self):
+        """
+        L{client.getPage} returns a L{Deferred} which is called back with
+        the body of the response if the default method B{GET} is used.
+        """
+        d = client.getPage(self.getURL("file"))
+        d.addCallback(self.assertEquals, "0123456789")
         return d
 
-    def testTimeoutTriggering(self):
-        # Test that when the timeout does trigger, we get a defer.TimeoutError.
-        return self.assertFailure(
-            client.getPage(self.getURL("wait"), timeout=0.5),
+
+    def test_getPageHead(self):
+        """
+        L{client.getPage} returns a L{Deferred} which is called back with
+        the empty string if the method is C{HEAD} and there is a successful
+        response code.
+        """
+        def getPage(method):
+            return client.getPage(self.getURL("file"), method=method)
+        return defer.gatherResults([
+            getPage("head").addCallback(self.assertEqual, ""),
+            getPage("HEAD").addCallback(self.assertEqual, "")])
+
+
+    def test_timeoutNotTriggering(self):
+        """
+        When a non-zero timeout is passed to L{getPage} and the page is
+        retrieved before the timeout period elapses, the L{Deferred} is
+        called back with the contents of the page.
+        """
+        d = client.getPage(self.getURL("host"), timeout=100)
+        d.addCallback(self.assertEquals, "127.0.0.1")
+        return d
+
+
+    def test_timeoutTriggering(self):
+        """
+        When a non-zero timeout is passed to L{getPage} and that many
+        seconds elapse before the server responds to the request. the
+        L{Deferred} is errbacked with a L{error.TimeoutError}.
+        """
+        finished = self.assertFailure(
+            client.getPage(self.getURL("wait"), timeout=0.000001),
             defer.TimeoutError)
+        def cleanup(passthrough):
+            # Clean up the server which is hanging around not doing
+            # anything.
+            connected = self.wrapper.protocols.keys()
+            # There might be nothing here if the server managed to already see
+            # that the connection was lost.
+            if connected:
+                connected[0].transport.loseConnection()
+            return passthrough
+        finished.addBoth(cleanup)
+        return finished
+
 
     def testDownloadPage(self):
         downloads = []
@@ -149,11 +302,11 @@ class WebClientTestCase(unittest.TestCase):
             d.addCallback(self._cbDownloadPageTest, data, name)
             downloads.append(d)
         return defer.gatherResults(downloads)
-    
+
     def _cbDownloadPageTest(self, ignored, data, name):
         bytes = file(name, "rb").read()
         self.assertEquals(bytes, data)
- 
+
     def testDownloadPageError1(self):
         class errorfile:
             def write(self, data):
@@ -208,24 +361,24 @@ class WebClientTestCase(unittest.TestCase):
 
     def testDownloadServerError(self):
         return self._downloadTest(lambda url: client.downloadPage(self.getURL(url), url.split('?')[0]))
-        
+
     def testFactoryInfo(self):
         url = self.getURL('file')
         scheme, host, port, path = client._parse(url)
         factory = client.HTTPClientFactory(url)
         reactor.connectTCP(host, port, factory)
         return factory.deferred.addCallback(self._cbFactoryInfo, factory)
-    
+
     def _cbFactoryInfo(self, ignoredResult, factory):
         self.assertEquals(factory.status, '200')
         self.assert_(factory.version.startswith('HTTP/'))
         self.assertEquals(factory.message, 'OK')
         self.assertEquals(factory.response_headers['content-length'][0], '10')
-        
+
 
     def testRedirect(self):
         return client.getPage(self.getURL("redirect")).addCallback(self._cbRedirect)
-    
+
     def _cbRedirect(self, pageData):
         self.assertEquals(pageData, "0123456789")
         d = self.assertFailure(
@@ -233,30 +386,49 @@ class WebClientTestCase(unittest.TestCase):
             error.PageRedirect)
         d.addCallback(self._cbCheckLocation)
         return d
-    
+
     def _cbCheckLocation(self, exc):
         self.assertEquals(exc.location, "/file")
 
+
+    def test_infiniteRedirection(self):
+        """
+        When more than C{redirectLimit} HTTP redirects are encountered, the
+        page request fails with L{InfiniteRedirection}.
+        """
+        def checkRedirectCount(*a):
+            self.assertEquals(f._redirectCount, 20)
+            self.assertEquals(self.infiniteRedirectResource.count, 20)
+
+        f = client._makeGetterFactory(
+            self.getURL('infiniteRedirect'),
+            client.HTTPClientFactory,
+            redirectLimit=20)
+        d = self.assertFailure(f.deferred, error.InfiniteRedirection)
+        d.addCallback(checkRedirectCount)
+        return d
+
+
     def testPartial(self):
         name = self.mktemp()
         f = open(name, "wb")
         f.write("abcd")
         f.close()
-        
+
         downloads = []
         partialDownload = [(True, "abcd456789"),
                            (True, "abcd456789"),
                            (False, "0123456789")]
-        
+
         d = defer.succeed(None)
         for (partial, expectedData) in partialDownload:
             d.addCallback(self._cbRunPartial, name, partial)
             d.addCallback(self._cbPartialTest, expectedData, name)
-        
+
         return d
 
     testPartial.skip = "Cannot test until webserver can serve partial data properly"
-    
+
     def _cbRunPartial(self, ignored, name, partial):
         return client.downloadPage(self.getURL("file"), name, supportPartial=partial)
 
@@ -269,8 +441,8 @@ class WebClientSSLTestCase(WebClientTestCase):
         from twisted import test
         return reactor.listenSSL(0, site,
                                  contextFactory=ssl.DefaultOpenSSLContextFactory(
-            sibpath(test.__file__, 'server.pem'),
-            sibpath(test.__file__, 'server.pem'),
+            FilePath(test.__file__).sibling('server.pem').path,
+            FilePath(test.__file__).sibling('server.pem').path,
             ),
                                  interface="127.0.0.1")
 
@@ -302,8 +474,8 @@ class WebClientRedirectBetweenSSLandPlainText(unittest.TestCase):
         from twisted import test
         self.tlsPort = reactor.listenSSL(0, tlsSite,
                                          contextFactory=ssl.DefaultOpenSSLContextFactory(
-            sibpath(test.__file__, 'server.pem'),
-            sibpath(test.__file__, 'server.pem'),
+            FilePath(test.__file__).sibling('server.pem').path,
+            FilePath(test.__file__).sibling('server.pem').path,
             ),
                                          interface="127.0.0.1")
         self.plainPort = reactor.listenTCP(0, plainSite, interface="127.0.0.1")
@@ -410,4 +582,3 @@ if ssl is None or not hasattr(ssl, 'DefaultOpenSSLContextFactory'):
 if not interfaces.IReactorSSL(reactor, None):
     for case in [WebClientSSLTestCase, WebClientRedirectBetweenSSLandPlainText]:
         case.skip = "Reactor doesn't support SSL"
-
