@@ -1,22 +1,17 @@
 # -*- test-case-name: twisted.web.test.test_xml -*-
-#
-# Copyright (c) 2001-2004 Twisted Matrix Laboratories.
+# Copyright (c) 2001-2008 Twisted Matrix Laboratories.
 # See LICENSE for details.
 
-#
-
-"""Some fairly inadequate testcases for Twisted XML support."""
-
-from __future__ import nested_scopes
+"""
+Some fairly inadequate testcases for Twisted XML support.
+"""
 
 from twisted.trial.unittest import TestCase
-
 from twisted.web import sux
-
 from twisted.web import microdom
-
 from twisted.web import domhelpers
 
+
 class Sux0r(sux.XMLParser):
     def __init__(self):
         self.tokens = []
@@ -42,6 +37,49 @@ class SUXTest(TestCase):
 
 class MicroDOMTest(TestCase):
 
+    def test_leadingTextDropping(self):
+        """
+        Make sure that if there's no top-level node lenient-mode won't
+        drop leading text that's outside of any elements.
+        """
+        s = "Hi orders! <br>Well. <br>"
+        d = microdom.parseString(s, beExtremelyLenient=True)
+        self.assertEquals(d.firstChild().toxml(),
+                          '<html>Hi orders! <br />Well. <br /></html>')
+
+    def test_trailingTextDropping(self):
+        """
+        Ensure that no *trailing* text in a mal-formed
+        no-top-level-element document(s) will not be dropped.
+        """
+        s = "<br>Hi orders!"
+        d = microdom.parseString(s, beExtremelyLenient=True)
+        self.assertEquals(d.firstChild().toxml(),
+                          '<html><br />Hi orders!</html>')
+
+
+    def test_noTags(self):
+        """
+        A string with nothing that looks like a tag at all should just
+        be parsed as body text.
+        """
+        s = "Hi orders!"
+        d = microdom.parseString(s, beExtremelyLenient=True)
+        self.assertEquals(d.firstChild().toxml(),
+                          "<html>Hi orders!</html>")
+
+
+    def test_surroundingCrap(self):
+        """
+        If a document is surrounded by non-xml text, the text should
+        be remain in the XML.
+        """
+        s = "Hi<br> orders!"
+        d = microdom.parseString(s, beExtremelyLenient=True)
+        self.assertEquals(d.firstChild().toxml(),
+                          "<html>Hi<br /> orders!</html>")
+
+
     def testCaseSensitiveSoonCloser(self):
         s = """
               <HTML><BODY>
@@ -60,10 +98,34 @@ class MicroDOMTest(TestCase):
         n = domhelpers.gatherTextNodes(l[0],1).replace('&nbsp;',' ')
         self.assertEquals(n.find('insane'), -1)
 
+
+    def test_lenientParenting(self):
+        """
+        Test that C{parentNode} attributes are set to meaningful values when
+        we are parsing HTML that lacks a root node.
+        """
+        # Spare the rod, ruin the child.
+        s = "<br/><br/>"
+        d = microdom.parseString(s, beExtremelyLenient=1)
+        self.assertIdentical(d.documentElement,
+                             d.documentElement.firstChild().parentNode)
+
+
+    def test_lenientParentSingle(self):
+        """
+        Test that the C{parentNode} attribute is set to a meaningful value
+        when we parse an HTML document that has a non-Element root node.
+        """
+        s = "Hello"
+        d = microdom.parseString(s, beExtremelyLenient=1)
+        self.assertIdentical(d.documentElement,
+                             d.documentElement.firstChild().parentNode)
+
+
     def testUnEntities(self):
         s = """
                 <HTML>
-                    This HTML goes between Stupid <=CrAzY!=> Dumb. 
+                    This HTML goes between Stupid <=CrAzY!=> Dumb.
                 </HTML>
             """
         d = microdom.parseString(s, beExtremelyLenient=1)
@@ -126,21 +188,25 @@ alert("I hate you");
         d = microdom.parseString(s, beExtremelyLenient=1)
         self.assertEquals(d.firstChild().firstChild().firstChild().data,
                           "(foo < bar) and (bar > foo)")
-        self.assertEquals(d.firstChild().getElementsByTagName("script")[1].firstChild().data,
-                          "foo </scrip bar ")
+        self.assertEquals(
+            d.firstChild().getElementsByTagName("script")[1].firstChild().data,
+            "foo </scrip bar ")
 
     def testScriptLeniencyIntelligence(self):
         # if there is comment or CDATA in script, the autoquoting in bEL mode
         # should not happen
         s = """<script><!-- lalal --></script>"""
-        self.assertEquals(microdom.parseString(s, beExtremelyLenient=1).firstChild().toxml(), s)
+        self.assertEquals(
+            microdom.parseString(s, beExtremelyLenient=1).firstChild().toxml(), s)
         s = """<script><![CDATA[lalal]]></script>"""
-        self.assertEquals(microdom.parseString(s, beExtremelyLenient=1).firstChild().toxml(), s)
+        self.assertEquals(
+            microdom.parseString(s, beExtremelyLenient=1).firstChild().toxml(), s)
         s = """<script> // <![CDATA[
         lalal
         //]]></script>"""
-        self.assertEquals(microdom.parseString(s, beExtremelyLenient=1).firstChild().toxml(), s)
-        
+        self.assertEquals(
+            microdom.parseString(s, beExtremelyLenient=1).firstChild().toxml(), s)
+
     def testPreserveCase(self):
         s = '<eNcApSuLaTe><sUxor></sUxor><bOrk><w00T>TeXt</W00t></BoRk></EnCaPsUlAtE>'
         s2 = s.lower().replace('text', 'TeXt')
@@ -317,7 +383,8 @@ alert("I hate you");
         s2 = '<foo/>'
         d = microdom.parseString(s)
         d2 = microdom.parseString(s2)
-        self.assertEquals(d.doctype, 'foo PUBLIC "baz" "http://www.example.com/example.dtd"')
+        self.assertEquals(d.doctype,
+                          'foo PUBLIC "baz" "http://www.example.com/example.dtd"')
         self.assertEquals(d.toxml(), s)
         self.failIf(d.isEqualToDocument(d2))
         self.failUnless(d.documentElement.isEqualToNode(d2.documentElement))
@@ -325,7 +392,8 @@ alert("I hate you");
     samples = [("<img/>", "<img />"),
                ("<foo A='b'>x</foo>", '<foo A="b">x</foo>'),
                ("<foo><BAR /></foo>", "<foo><BAR></BAR></foo>"),
-               ("<foo>hello there &amp; yoyoy</foo>", "<foo>hello there &amp; yoyoy</foo>"),
+               ("<foo>hello there &amp; yoyoy</foo>",
+                "<foo>hello there &amp; yoyoy</foo>"),
                ]
 
     def testOutput(self):
@@ -376,15 +444,18 @@ alert("I hate you");
                      ("& ", "&amp; "),
                      ("&amp;", "&amp;"),
                      ("&hello monkey", "&amp;hello monkey")]:
-            d = microdom.parseString("%s<pre>%s</pre>" % (prefix, i), beExtremelyLenient=1)
+            d = microdom.parseString("%s<pre>%s</pre>"
+                                     % (prefix, i), beExtremelyLenient=1)
             self.assertEquals(d.documentElement.toxml(), "<pre>%s</pre>" % o)
         # non-space preserving
-        d = microdom.parseString("%s<t>hello & there</t>", beExtremelyLenient=1)
+        d = microdom.parseString("<t>hello & there</t>", beExtremelyLenient=1)
         self.assertEquals(d.documentElement.toxml(), "<t>hello &amp; there</t>")
-    
+
     def testInsensitiveLenient(self):
         # testing issue #537
-        d = microdom.parseString("<?xml version='1.0'?><bar><xA><y>c</Xa> <foo></bar>", beExtremelyLenient=1)
+        d = microdom.parseString(
+            "<?xml version='1.0'?><bar><xA><y>c</Xa> <foo></bar>",
+            beExtremelyLenient=1)
         self.assertEquals(d.documentElement.firstChild().toxml(), "<xa><y>c</y></xa>")
 
     def testSpacing(self):
@@ -425,9 +496,8 @@ alert("I hate you");
         d = microdom.parseString(s, beExtremelyLenient=1)
         actual = d.documentElement.toxml()
         self.assertEquals(expected, actual)
-        
     testLaterCloserTable.todo = "Table parsing needs to be fixed."
-    
+
     def testLaterCloserDL(self):
         s = ("<dl>"
              "<dt>word<dd>definition"
@@ -455,7 +525,7 @@ alert("I hate you");
         self.assertEquals(expected, actual)
 
     testLaterCloserDL2.todo = "unclosed <p> messes it up."
-    
+
     def testUnicodeTolerance(self):
         import struct
         s = '<foo><bar><baz /></bar></foo>'
@@ -520,9 +590,9 @@ alert("I hate you");
         self.assertEquals(node.namespace, clone.namespace)
 
     def testCloneDocument(self):
-        # sorry bout the >80 cols, but whitespace is a sensitive thing
-        s = '''<?xml version="1.0"?><!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
-"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"><foo></foo>'''
+        s = ('<?xml version="1.0"?>'
+             '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"'
+             '"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"><foo></foo>')
 
         node = microdom.parseString(s)
         clone = node.cloneNode(deep=1)
@@ -575,13 +645,322 @@ alert("I hate you");
                           "base")
         self.assertEquals(d.documentElement.getElementsByTagName("y")[0].namespace,
                           "base")
-        self.assertEquals(d.documentElement.getElementsByTagName("y")[1].getAttributeNS('base','q'),
-                          '1')
-        
+        self.assertEquals(
+            d.documentElement.getElementsByTagName("y")[1].getAttributeNS('base','q'),
+            '1')
+
         d2 = microdom.parseString(s2)
         self.assertEquals(d2.documentElement.namespace,
                           "base")
         self.assertEquals(d2.documentElement.getElementsByTagName("y")[0].namespace,
                           "base")
-        self.assertEquals(d2.documentElement.getElementsByTagName("y")[1].getAttributeNS('base','q'),
-                          '1')
+        self.assertEquals(
+            d2.documentElement.getElementsByTagName("y")[1].getAttributeNS('base','q'),
+            '1')
+
+    def testNamespaceDelete(self):
+        """
+        Test that C{toxml} can support xml structures that remove namespaces.
+        """
+        s1 = ('<?xml version="1.0"?><html xmlns="http://www.w3.org/TR/REC-html40">'
+              '<body xmlns=""></body></html>')
+        s2 = microdom.parseString(s1).toxml()
+        self.assertEquals(s1, s2)
+
+    def testNamespaceInheritance(self):
+        """
+        Check that unspecified namespace is a thing separate from undefined
+        namespace. This test added after discovering some weirdness in Lore.
+        """
+        # will only work if childNodes is mutated. not sure why.
+        child = microdom.Element('ol')
+        parent = microdom.Element('div', namespace='http://www.w3.org/1999/xhtml')
+        parent.childNodes = [child]
+        self.assertEquals(parent.toxml(),
+                          '<div xmlns="http://www.w3.org/1999/xhtml"><ol></ol></div>')
+
+
+class TestBrokenHTML(TestCase):
+    """
+    Tests for when microdom encounters very bad HTML and C{beExtremelyLenient}
+    is enabled. These tests are inspired by some HTML generated in by a mailer,
+    which breaks up very long lines by splitting them with '!\n '. The expected
+    behaviour is loosely modelled on the way Firefox treats very bad HTML.
+    """
+
+    def checkParsed(self, input, expected, beExtremelyLenient=1):
+        """
+        Check that C{input}, when parsed, produces a DOM where the XML
+        of the document element is equal to C{expected}.
+        """
+        output = microdom.parseString(input,
+                                      beExtremelyLenient=beExtremelyLenient)
+        self.assertEquals(output.documentElement.toxml(), expected)
+
+
+    def test_brokenAttributeName(self):
+        """
+        Check that microdom does its best to handle broken attribute names.
+        The important thing is that it doesn't raise an exception.
+        """
+        input = '<body><h1><div al!\n ign="center">Foo</div></h1></body>'
+        expected = ('<body><h1><div ign="center" al="True">'
+                    'Foo</div></h1></body>')
+        self.checkParsed(input, expected)
+
+
+    def test_brokenAttributeValue(self):
+        """
+        Check that microdom encompasses broken attribute values.
+        """
+        input = '<body><h1><div align="cen!\n ter">Foo</div></h1></body>'
+        expected = '<body><h1><div align="cen!\n ter">Foo</div></h1></body>'
+        self.checkParsed(input, expected)
+
+
+    def test_brokenOpeningTag(self):
+        """
+        Check that microdom does its best to handle broken opening tags.
+        The important thing is that it doesn't raise an exception.
+        """
+        input = '<body><h1><sp!\n an>Hello World!</span></h1></body>'
+        expected = '<body><h1><sp an="True">Hello World!</sp></h1></body>'
+        self.checkParsed(input, expected)
+
+
+    def test_brokenSelfClosingTag(self):
+        """
+        Check that microdom does its best to handle broken self-closing tags
+        The important thing is that it doesn't raise an exception.
+        """
+        self.checkParsed('<body><span /!\n></body>',
+                         '<body><span></span></body>')
+        self.checkParsed('<span!\n />', '<span></span>')
+
+
+    def test_brokenClosingTag(self):
+        """
+        Check that microdom does its best to handle broken closing tags.
+        The important thing is that it doesn't raise an exception.
+        """
+        input = '<body><h1><span>Hello World!</sp!\nan></h1></body>'
+        expected = '<body><h1><span>Hello World!</span></h1></body>'
+        self.checkParsed(input, expected)
+        input = '<body><h1><span>Hello World!</!\nspan></h1></body>'
+        self.checkParsed(input, expected)
+        input = '<body><h1><span>Hello World!</span!\n></h1></body>'
+        self.checkParsed(input, expected)
+        input = '<body><h1><span>Hello World!<!\n/span></h1></body>'
+        expected = '<body><h1><span>Hello World!<!></!></span></h1></body>'
+        self.checkParsed(input, expected)
+
+
+
+
+class NodeTests(TestCase):
+    """
+    Tests for L{Node}.
+    """
+    def test_isNodeEqualTo(self):
+        """
+        L{Node.isEqualToNode} returns C{True} if and only if passed a L{Node}
+        with the same children.
+        """
+        # A node is equal to itself
+        node = microdom.Node(object())
+        self.assertTrue(node.isEqualToNode(node))
+        another = microdom.Node(object())
+        # Two nodes with no children are equal
+        self.assertTrue(node.isEqualToNode(another))
+        node.appendChild(microdom.Node(object()))
+        # A node with no children is not equal to a node with a child
+        self.assertFalse(node.isEqualToNode(another))
+        another.appendChild(microdom.Node(object()))
+        # A node with a child and no grandchildren is equal to another node
+        # with a child and no grandchildren.
+        self.assertTrue(node.isEqualToNode(another))
+        # A node with a child and a grandchild is not equal to another node
+        # with a child and no grandchildren.
+        node.firstChild().appendChild(microdom.Node(object()))
+        self.assertFalse(node.isEqualToNode(another))
+        # A node with a child and a grandchild is equal to another node with a
+        # child and a grandchild.
+        another.firstChild().appendChild(microdom.Node(object()))
+        self.assertTrue(node.isEqualToNode(another))
+
+
+
+class DocumentTests(TestCase):
+    """
+    Tests for L{Document}.
+    """
+    doctype = 'foo PUBLIC "baz" "http://www.example.com/example.dtd"'
+
+    def test_isEqualToNode(self):
+        """
+        L{Document.isEqualToNode} returns C{True} if and only if passed a
+        L{Document} with the same C{doctype} and C{documentElement}.
+        """
+        # A document is equal to itself
+        document = microdom.Document()
+        self.assertTrue(document.isEqualToNode(document))
+        # A document without a doctype or documentElement is equal to another
+        # document without a doctype or documentElement.
+        another = microdom.Document()
+        self.assertTrue(document.isEqualToNode(another))
+        # A document with a doctype is not equal to a document without a
+        # doctype.
+        document.doctype = self.doctype
+        self.assertFalse(document.isEqualToNode(another))
+        # Two documents with the same doctype are equal
+        another.doctype = self.doctype
+        self.assertTrue(document.isEqualToNode(another))
+        # A document with a documentElement is not equal to a document without
+        # a documentElement
+        document.appendChild(microdom.Node(object()))
+        self.assertFalse(document.isEqualToNode(another))
+        # Two documents with equal documentElements are equal.
+        another.appendChild(microdom.Node(object()))
+        self.assertTrue(document.isEqualToNode(another))
+        # Two documents with documentElements which are not equal are not
+        # equal.
+        document.documentElement.appendChild(microdom.Node(object()))
+        self.assertFalse(document.isEqualToNode(another))
+
+
+
+class EntityReferenceTests(TestCase):
+    """
+    Tests for L{EntityReference}.
+    """
+    def test_isEqualToNode(self):
+        """
+        L{EntityReference.isEqualToNode} returns C{True} if and only if passed
+        a L{EntityReference} with the same C{eref}.
+        """
+        self.assertTrue(
+            microdom.EntityReference('quot').isEqualToNode(
+                microdom.EntityReference('quot')))
+        self.assertFalse(
+            microdom.EntityReference('quot').isEqualToNode(
+                microdom.EntityReference('apos')))
+
+
+
+class CharacterDataTests(TestCase):
+    """
+    Tests for L{CharacterData}.
+    """
+    def test_isEqualToNode(self):
+        """
+        L{CharacterData.isEqualToNode} returns C{True} if and only if passed a
+        L{CharacterData} with the same value.
+        """
+        self.assertTrue(
+            microdom.CharacterData('foo').isEqualToNode(
+                microdom.CharacterData('foo')))
+        self.assertFalse(
+            microdom.CharacterData('foo').isEqualToNode(
+                microdom.CharacterData('bar')))
+
+
+
+class CommentTests(TestCase):
+    """
+    Tests for L{Comment}.
+    """
+    def test_isEqualToNode(self):
+        """
+        L{Comment.isEqualToNode} returns C{True} if and only if passed a
+        L{Comment} with the same value.
+        """
+        self.assertTrue(
+            microdom.Comment('foo').isEqualToNode(
+                microdom.Comment('foo')))
+        self.assertFalse(
+            microdom.Comment('foo').isEqualToNode(
+                microdom.Comment('bar')))
+
+
+
+class TextTests(TestCase):
+    """
+    Tests for L{Text}.
+    """
+    def test_isEqualToNode(self):
+        """
+        L{Text.isEqualToNode} returns C{True} if and only if passed a L{Text}
+        which represents the same data.
+        """
+        self.assertTrue(
+            microdom.Text('foo', raw=True).isEqualToNode(
+                microdom.Text('foo', raw=True)))
+        self.assertFalse(
+            microdom.Text('foo', raw=True).isEqualToNode(
+                microdom.Text('foo', raw=False)))
+        self.assertFalse(
+            microdom.Text('foo', raw=True).isEqualToNode(
+                microdom.Text('bar', raw=True)))
+
+
+
+class CDATASectionTests(TestCase):
+    """
+    Tests for L{CDATASection}.
+    """
+    def test_isEqualToNode(self):
+        """
+        L{CDATASection.isEqualToNode} returns C{True} if and only if passed a
+        L{CDATASection} which represents the same data.
+        """
+        self.assertTrue(
+            microdom.CDATASection('foo').isEqualToNode(
+                microdom.CDATASection('foo')))
+        self.assertFalse(
+            microdom.CDATASection('foo').isEqualToNode(
+                microdom.CDATASection('bar')))
+
+
+
+class ElementTests(TestCase):
+    """
+    Tests for L{Element}.
+    """
+    def test_isEqualToNode(self):
+        """
+        L{Element.isEqualToNode} returns C{True} if and only if passed a
+        L{Element} with the same C{nodeName}, C{namespace}, C{childNodes}, and
+        C{attributes}.
+        """
+        self.assertTrue(
+            microdom.Element(
+                'foo', {'a': 'b'}, object(), namespace='bar').isEqualToNode(
+                microdom.Element(
+                    'foo', {'a': 'b'}, object(), namespace='bar')))
+
+        # Elements with different nodeName values do not compare equal.
+        self.assertFalse(
+            microdom.Element(
+                'foo', {'a': 'b'}, object(), namespace='bar').isEqualToNode(
+                microdom.Element(
+                    'bar', {'a': 'b'}, object(), namespace='bar')))
+
+        # Elements with different namespaces do not compare equal.
+        self.assertFalse(
+            microdom.Element(
+                'foo', {'a': 'b'}, object(), namespace='bar').isEqualToNode(
+                microdom.Element(
+                    'foo', {'a': 'b'}, object(), namespace='baz')))
+
+        # Elements with different childNodes do not compare equal.
+        one = microdom.Element('foo', {'a': 'b'}, object(), namespace='bar')
+        two = microdom.Element('foo', {'a': 'b'}, object(), namespace='bar')
+        two.appendChild(microdom.Node(object()))
+        self.assertFalse(one.isEqualToNode(two))
+
+        # Elements with different attributes do not compare equal.
+        self.assertFalse(
+            microdom.Element(
+                'foo', {'a': 'b'}, object(), namespace='bar').isEqualToNode(
+                microdom.Element(
+                    'foo', {'a': 'c'}, object(), namespace='bar')))
